# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Reader to read GCOM-C1 SGLI files.

(NRL-Monterey, Oct. 2025)
"""

# Python Standard Libraries
import logging
from os.path import basename
from datetime import datetime

# Installed libraries
import xarray as xr
import numpy as np
from scipy.ndimage import zoom
from h5py import File as h5File

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "sgli_l1b_hdf5"

"""
Channel Info:
channel name, center wavelength (um), resolution (km)
SW01, 1.05, 1
SW02, 1.38, 1
SW03, 1.63, .25
SW04, 2.21, 1
TI01, 10.8, .5
TI02, 12, .5

VN01, .38, .25
VN02, .412, .25
VN03, .443, .25
VN04, .49, .25
VN05, .53, .25
VN06, .565, .25
VN07, .6735, .25
VN08, .6735, .25
VN09, .763, 1
VN10, .8685, .25
VN11, .8685, .25

P01, .6735, 1
P02, .8685, 1
(Pol-chans are seperated by stokes polarization: I,Q,U)
"""

ir_chans = {"Lt_SW01", "Lt_SW02", "Lt_SW03", "Lt_SW04", "Lt_TI01", "Lt_TI02"}
vis_chans = {
    "Lt_VN01",
    "Lt_VN02",
    "Lt_VN03",
    "Lt_VN04",
    "Lt_VN05",
    "Lt_VN06",
    "Lt_VN07",
    "Lt_VN08",
    "Lt_VN09",
    "Lt_VN10",
    "Lt_VN11",
}
pol_chans = {"Lt_PI01", "Lt_PI02", "Lt_PQ01", "Lt_PQ02", "Lt_PU01", "Lt_PU02"}

res_mapping = {1000: "LOW", 500: "MED", 250: "HIGH"}


def commit_to_dict(data_dict, key, tmp_xry):
    """Commit data as a list item to a dictionary."""
    if key in data_dict.keys():
        data_dict[key].append(tmp_xry)
    else:
        data_dict[key] = [tmp_xry]

    return data_dict


def rad_to_temp(rad, center_wl):
    """Convert Radiance to Temp."""
    c1 = 1.1911042e8
    c2 = 1.4387752e4
    temp = c2 / (center_wl * np.log(c1 / ((center_wl**5) * rad + 1)))

    return temp


def read_irs(infile, data_chans=ir_chans):
    """Read the infrared scanner (SWIR & TIR) data."""
    # three possible resolutions
    # HIGH (250m) MED (500m) LOW (1km)
    xrt = xr.Dataset(attrs={"res_types": []})
    with h5File(infile) as irs_h5:
        # load all respective channels
        for k in data_chans:
            raw_chan = irs_h5["Image_data"][k]
            # mask data, top two bits are masked
            chan_name = k.split("_")[1]
            raw_img = raw_chan[...] & 0x3FFF
            raw_ma = np.where(raw_img == 16383, np.nan, raw_img)
            # create radiance, reflectance
            toa_rad = (raw_chan.attrs["Slope"][...] * raw_ma) + raw_chan.attrs[
                "Offset"
            ][...]
            ref_flag = False
            if "Slope_reflectance" in raw_chan.attrs.keys():
                toa_ref = (
                    raw_chan.attrs["Slope_reflectance"][...] * raw_ma
                ) + raw_chan.attrs["Offset_reflectance"][...]
                ref_flag = True

            res_type = res_mapping[int(raw_chan.attrs["Spatial_resolution"][0])]
            if res_type not in xrt.attrs["res_types"]:
                # commit geo values
                xrt.attrs["res_types"].append(res_type)
                # load geo channels
                lat = irs_h5["Geometry_data"]["Latitude"][...]
                lon = irs_h5["Geometry_data"]["Longitude"][...]
                sol_azi_angle = irs_h5["Geometry_data"]["Solar_azimuth"][...] * 0.01
                sol_zen_angle = irs_h5["Geometry_data"]["Solar_zenith"][...] * 0.01
                sat_azi = irs_h5["Geometry_data"]["Sensor_azimuth"][...] * 0.01
                sat_zen = irs_h5["Geometry_data"]["Sensor_zenith"][...] * 0.01
                # zoom if needed
                if lat.shape != raw_img.shape:
                    coef = (
                        raw_img.shape[0] / lat.shape[0],
                        raw_img.shape[1] / lat.shape[1],
                    )
                    lat = zoom(lat, coef)
                    lon = zoom(lon, coef)
                    sat_azi = zoom(sat_azi, coef)
                    sat_zen = zoom(sat_zen, coef)
                    sol_zen_angle = zoom(sol_zen_angle, coef)
                    sol_azi_angle = zoom(sol_azi_angle, coef)

                xrt[res_type.lower() + "lat"] = xr.DataArray(
                    lat.ravel(), dims=(res_type.lower() + "_dim0")
                )
                xrt[res_type.lower() + "lon"] = xr.DataArray(
                    lon.ravel(), dims=(res_type.lower() + "_dim0")
                )
                xrt[res_type.lower() + "sat_zen"] = xr.DataArray(
                    sat_zen.ravel(), dims=(res_type.lower() + "_dim0")
                )
                xrt[res_type.lower() + "sat_azi"] = xr.DataArray(
                    sat_azi.ravel(), dims=(res_type.lower() + "_dim0")
                )
                xrt[res_type.lower() + "sol_zen"] = xr.DataArray(
                    sol_zen_angle.ravel(), dims=(res_type.lower() + "_dim0")
                )
                xrt[res_type.lower() + "sol_azi"] = xr.DataArray(
                    sol_azi_angle.ravel(), dims=(res_type.lower() + "_dim0")
                )
            # commit data to xarray
            xrt[chan_name + "Rad"] = xr.DataArray(
                toa_rad.ravel(), dims=(res_type.lower() + "_dim0",)
            )
            temp = rad_to_temp(toa_rad, raw_chan.attrs["Center_wavelength"][...] / 1000)
            xrt[chan_name + "BT"] = xr.DataArray(
                temp.ravel(), dims=(res_type.lower() + "_dim0",)
            )

            if ref_flag:
                xrt[chan_name + "Ref"] = xr.DataArray(
                    toa_ref.ravel(),
                    dims=(res_type.lower() + "_dim0",),
                )

    return xrt


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
):
    """Read GFS GRIB data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    base_fnames = list(map(basename, fnames))
    fnames = sorted(fnames)

    # approximate for metadata only
    dtime_start = datetime.strptime(base_fnames[0].split("_")[1][:10], "%Y%m%d%H%M")
    dtime_end = datetime.strptime(base_fnames[-1].split("_")[1][:10], "%Y%m%d%H%M")

    # if mixed res assume low
    res_types = set(f.split("_")[3][-1] for f in base_fnames)
    res_dict = {"K": 1, "L": 1, "Q": 0.25}

    if len(list(res_types)) > 1:
        res = 1
    else:
        res = res_dict[res_types.pop()]

    tmp_xr = xr.Dataset(
        attrs={
            "source_file_name": base_fnames,
            "start_datetime": dtime_start,
            "end_datetime": dtime_end,
            "source_name": "sgli",
            "platform_name": "GCOM-C1",
            "data_provider": "JAXA",
            "sample_distance_km": res,
            "interpolation_radius_of_influence": 3000,
        }
    )
    if metadata_only:
        tmp_dict = {"METADATA": tmp_xr}
        return tmp_dict

    vnr_chans = vis_chans
    irs_chans = ir_chans

    if chans is not None:
        req_chans = set("Lt_" + i[:4] for i in chans)
        irs_chans = req_chans & ir_chans
        vnr_chans = req_chans & vis_chans
        polo_chans = req_chans & pol_chans
        if (
            len(list(irs_chans)) == 0
            and len(list(vnr_chans)) == 0
            and len(list(polo_chans)) == 0
        ):
            raise ValueError(
                "No channels found for SGLI, check requestd channels: {}".format(
                    req_chans
                )
            )
            # read data, seperated by resolution
    data_xr_dict = {}
    res_list = []
    for f in fnames:
        f_basename = basename(f)
        if "IRS" in f_basename:
            data_xarray = read_irs(f, data_chans=irs_chans)
            data_xr_dict = commit_to_dict(data_xr_dict, "IRS", data_xarray)
        elif "VNR" in f_basename:
            data_xarray = read_irs(f, data_chans=vnr_chans)
            data_xr_dict = commit_to_dict(data_xr_dict, "VNR", data_xarray)
        elif "POL" in f_basename:
            data_xarray = read_irs(f, data_chans=polo_chans)
            data_xr_dict = commit_to_dict(data_xr_dict, "POL", data_xarray)
        else:
            LOG.warning("No data type readable in file: {}".format(f_basename))
            raise ValueError("No matching file key for IRS, VNR, or POL in filename.")

        res_list.extend(data_xarray.attrs["res_types"])

    if len(list(data_xr_dict.keys())) == 1:
        single_key = list(data_xr_dict.keys())[0]
        full_xarray = xr.concat(data_xr_dict[single_key], dim="dimN")
    else:
        # multi-file merge/concat
        for k in data_xr_dict.keys():
            data_xr_dict[k] = xr.concat(data_xr_dict[k], dim="dimN")
        full_xarray = xr.merge(data_xr_dict.values())

    final_data = {}
    # Reorder science data for a dict based upon resolution
    res_avail = set(res_list)
    for r in set(res_list):

        res_pop = res_avail - {r}
        res_pop = [i.lower() + "_dim0" for i in res_pop]
        final_data[r] = full_xarray.drop_dims(res_pop).rename(
            {
                r.lower() + "lat": "latitude",
                r.lower() + "lon": "longitude",
                r.lower() + "_dim0": "dim_0",
                r.lower() + "sat_zen": "satellite_zenith_angle",
                r.lower() + "sat_azi": "satellite_azimuth_angle",
                r.lower() + "sol_zen": "solar_zenith_angle",
                r.lower() + "sol_azi": "solar_azimuth_angle",
            }
        )
        final_data[r].attrs["res_types"] = r
        final_data[r].attrs |= tmp_xr.attrs

    final_data |= {"METADATA": tmp_xr}

    return final_data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Sector interface module."""

from cartopy import feature as cfeature
import numpy as np
from pyresample import kd_tree

from geoips.interfaces.base import BaseYamlPlugin, BaseYamlInterface
from geoips.image_utils.mpl_utils import create_figure_and_main_ax_and_mapobj

# Uncomment when ready to switch from JsonSchema to Pydantic
# from geoips.pydantic.sectors import SectorPluginModel

# Commenting these out for PR #260
# Will work on this again after the 2023 workshop
#
# def center_to_area_definition(sector):
#     """Return a pyresample AreaDefinition for the input sector.
#
#     The input sector must supply location information in the center format.
#     """
#     if not sector.family.startswith("center"):
#         raise ValueError("Sector does not supply location as center coordinates.")
#     raise NotImplementedError
#
#
# def corners_to_area_definition(sector):
#     """Return a pyresample AreaDefinition for the input sector.
#
#     The input sector must supply location information in the "corners" format.
#     """
#     if not sector.family.startswith("center"):
#         raise ValueError("Sector does not supply location as corner coordinates.")
#
#     ad_info = {
#         "area_id": sector.name,
#         "projection": {
#             "units": "m",
#             "a": 6371228.0,
#             "proj": sector.spec.projection,
#             "lat_0": sector.spec.center.lat,
#             "lon_0": sector.spec.center.lon,
#         },
#         "width": sector.shape[0],
#         "height": sector.shape[1],
#         "resolution": [sector.resolution, sector.resolution],
#         "center": [0, 0],
#     }
#
#     ad = create_area_def(**ad_info)
#     raise NotImplementedError


class SectorPluginBase(BaseYamlPlugin):
    """The base class for all sector plugins.

    This class provides the functionality specific to sector plugin classes. It should
    not be instantiated directly. To access sector plugins, use
    `geoips.interfaces.sectors`.
    """

    @property
    def area_definition(self):
        """Return the pyresample AreaDefinition for the sector."""
        # if self.family.startswith(("area_definition", "generated")):
        from pyresample import load_area
        from importlib.resources import files

        if self.family.startswith("area_definition"):
            abspath = str(files(self.package) / self.relpath)
            ad = load_area(abspath, "spec")
        # elif self.family.startswith("center"):
        #     ad = center_to_area_definition(self)
        # elif self.family.startswith("corners"):
        #     ad = corners_to_area_definition(self)
        return ad

    def create_test_plot(self, fname, return_fig_ax_map=False, overlay=False):
        """Create a test PNG image for this sector.

        Parameters
        ----------
        fname: str
            - The full path to the output image.
        return_fix_ax_map: bool, default=False
            - Whether or not we should save the image to disk or return the figure,
              axes, and mapobj variables in memory
        overlay: bool, default=False
            - If true, overlay this sector on the global grid and make it slightly
              transparent. Useful for projecting tiny sectors on the global grid to get
              a sense of where they'll end up and what they'll look like.

        """
        if overlay:
            global_sector = sectors.get_plugin("global_cylindrical")
            global_area_def = global_sector.area_definition
            fig, ax, mapobj = create_figure_and_main_ax_and_mapobj(
                global_area_def.shape[1],
                global_area_def.shape[0],
                global_area_def,
                noborder=True,
            )

            # Create a dummy 2D numpy array of data for self.area_definition
            data_overlay = np.ones(self.area_definition.shape) * 10  # or real data

            # Reproject data_overlay to the global grid
            resampled_data = kd_tree.resample_nearest(
                self.area_definition,
                data_overlay,
                global_area_def,
                radius_of_influence=14000,  # Not sure what to set here. I chose 14km
                fill_value=np.nan,
            )

            # Normalize to 0–1 range; this additionally masks out pixels we don't want
            # to color.
            norm = np.nan_to_num(resampled_data)
            norm = (norm - norm.min()) / (norm.max() - norm.min() + 1e-6)

            data_overlay = np.zeros(resampled_data.shape + (4,), dtype=np.float32)
            # This produces a Vivid Green Color
            data_overlay[..., 0] = 0.282
            data_overlay[..., 1] = 1.0
            data_overlay[..., 2] = 0
            data_overlay[..., 3] = norm * 0.5

            # Plot overlay data (with transparency) on the global grid
            ax.imshow(
                data_overlay,
                transform=mapobj,
                extent=(
                    global_area_def.area_extent[0],
                    global_area_def.area_extent[2],
                    global_area_def.area_extent[1],
                    global_area_def.area_extent[3],
                ),
                origin="upper",
            )

        else:
            fig, ax, mapobj = create_figure_and_main_ax_and_mapobj(
                self.area_definition.shape[1],
                self.area_definition.shape[0],
                self.area_definition,
                noborder=True,
            )
        ax.add_feature(cfeature.COASTLINE)
        ax.add_feature(cfeature.BORDERS)
        if fname is not None:
            fig.savefig(fname)
        if return_fig_ax_map:
            return fig, ax, mapobj


class SectorsInterface(BaseYamlInterface):
    """Interface for sector plugins."""

    name = "sectors"
    plugin_class = SectorPluginBase
    # Uncomment when ready to switch from JsonSchema to Pydantic
    # validator = SectorPluginModel


sectors = SectorsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Pytest file for calling integration bash scripts."""

import os
import pytest

from tests.integration_tests.test_integration import base_setup  # noqa: F401

from tests.integration_tests.test_integration import (
    run_script_with_bash,
    setup_environment as setup_geoips_environment,
)

global_example_integ_test_calls = [
    ##################################################################################
    # ### Scripts to produce global imagery outputs specifically for diagnostic
    # ### purposes, and identifying day/night/spatial contents of a given dataset.
    # ###
    # ### The below script will automatically run the following products
    # ###   * Test-*-Day-Night
    # ###   * Test-*-Day-Only
    # ###   * Test-*-Night-Only
    # ###
    # ### With satellite zenith cutoff of
    # ###   * 70
    # ###   * null
    # ###
    # ### Reprojected to a global domain in an annotated imagery output.
    # ###
    # ### Available product names can be found in
    # ### geoips/plugins/yaml/products/integration_tests/Test-Day-Night.yaml
    # ### geoips/plugins/yaml/products/integration_tests/Test-Day-Only.yaml
    # ### geoips/plugins/yaml/products/integration_tests/Test-Night-Only.yaml
    # ###
    # ### Product prefix should be passed below as Test-SENSOR-VAR
    ##################################################################################
    # MSG-3, METEOSAT-10, METEO-EU dataset, this dataset is nearly all day.
    # Note we have no currently available METEO-IO dataset (MSG-1 or 2)
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "seviri_hrit "
        "Test-SEVIRI-B9 "
        "$GEOIPS_TESTDATA_DIR/test_data_seviri/data/20250624/1200/H-000-MSG3* "
    ),
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "ami_netcdf "
        "Test-AMI-11p2 "
        "$GEOIPS_TESTDATA_DIR/test_data_ami/data/20231208_0300_daytime/gk2a_ami_le1b_*.nc"  # noqa: E501
    ),
    # MTG-I1, METEOSAT-12, METEO-EU dataset, this dataset is nearly all day.
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "fci_netcdf "
        "Test-FCI-B14 "
        "$GEOIPS_TESTDATA_DIR/test_data_fci/data/20250623/1200/W_XX-*.nc"
    ),
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "ahi_hsd "
        "Test-AHI-B13 "
        "$GEOIPS_TESTDATA_DIR/test_data_ahi/data/20200405_0000/HS_H08*"  # noqa: E501
    ),
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "abi_netcdf "
        "Test-ABI-B14 "
        "${GEOIPS_TESTDATA_DIR}/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*_G16_*.nc"  # noqa: E501
    ),
    (
        "$geoips_repopath/tests/example_scripts/global_terminator_satzen.sh "
        "abi_netcdf "
        "Test-ABI-B14 "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*_G17_*.nc"  # noqa: E501
    ),
]

tiny_sector_overlay_integ_test_calls = [
    ##################################################################################
    # ### Output geoips test sector overlays on a global map to identify location of
    # ### each defined tiny sector.  Used for diagnostic purposes to evaluate location
    # ### and size of each tiny sector.
    ##################################################################################
    ##################################################################################
    # ### Satellite specific test sectors, not reliant on a specific dataset
    ##################################################################################
    "geoips test sector --overlay test_goeswest_eqc_3km_nadir -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goeswest_eqc_3km_landocean -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goeswest_longlat_50km_fulldisk -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goeseast_eqc_3km_nadir -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goeseast_eqc_3km_landocean -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goeseast_longlat_50km_fulldisk -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_himawari_eqc_3km_landocean -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_meteoeu_eqc_3km_landocean -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_meteoeu_longlat_50km_fulldisk -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_meteoio_eqc_3km_landocean -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_meteoio_longlat_50km_fulldisk -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    ##################################################################################
    # ### Dataset specific test sectors, reliant on specific time of day and/or
    # ### meteorological features within the dataset.
    ##################################################################################
    "geoips test sector --overlay test_goes16_eqc_10km_edge_day_20200918T1950Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes16_eqc_10km_edge_night_20200918T1950Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes16_eqc_3km_day_20200918T1950Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes16_eqc_3km_night_20200918T1950Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes16_eqc_3km_terminator_20200918T1950Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes17_eqc_10km_edge_day_20210718T0150Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes17_eqc_10km_edge_night_20210718T0150Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes17_eqc_3km_day_20210718T0150Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes17_eqc_3km_night_20210718T0150Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_goes17_eqc_3km_terminator_20210718T0150Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
    "geoips test sector --overlay test_himawari8_eqc_10km_terminator_20200405T0000Z -o $GEOIPS_OUTDIRS/example_test_imagery_outputs ",  # noqa: E501
]

tiny_sector_integ_test_calls = [
    ##################################################################################
    # ### Test the actual output of each specified tiny sector defined within this repo.
    # ### tiny_sectors_geostationary.sh is a wrapper test script to produce the
    # ### specified tiny sector output, and compare it with the stored comparison
    # ### output imagery.
    # ###
    # ### It always produces
    # ### * imagery_clean output
    # ### * geoips_fname filename formatter,
    # ### * --compare_path "$GEOIPS_PACKAGES_DIR/${repo_name}/tests/integration_tests/tiny_sectors/outputs/${test_sector_name}_${product_name}"   # noqa: E501
    # ### * satellite zenith angle cutoff None,
    # ###
    # ### with the following arguments passed in:
    # ###
    # ### * test_sector_name
    # ###   * Usually of format test_SATELLITE_PROJ_RES_FEAT_DAYNIGHT_YYYYMMDDTHHMNZ
    # ###   * e.g. test_goes16_eqc_3km_edge_day_20200918T1950Z
    # ###   * e.g. test_goeswest_eqc_3km_nadir
    # ### * reader_name
    # ### * product_name
    # ###   * Test-*-Day-Only, or -Night-Only for efficiency
    # ###     reasons, consistency, and quickly identifying
    # ###     the terminator location
    # ### * repo_name
    # ### * data_files
    # ###
    # ### Available product names of the format Test-SENSOR-Var-* can be found in
    # ### geoips/plugins/yaml/products/integration_tests/Test-Day-Night.yaml
    # ### geoips/plugins/yaml/products/integration_tests/Test-Day-Only.yaml
    # ### geoips/plugins/yaml/products/integration_tests/Test-Night-Only.yaml
    # ###
    # ### NOTE if you set the environment variable
    # ### export GEOIPS_CREATE_TINY_SECTOR_TEST_GEOTIFF_OUTPUTS=True
    # ### then tiny_sectors_geostationary.sh will produce a geotiff output (with
    # ### no image output comparison) for evaluation / reference purposes, in addition
    # ### to the imagery_clean output that is used in the required integration tests.
    # ### This environment variable can be used for evaluation/reference purposes.
    ##################################################################################
    ##################################################################################
    # ### Satellite specific test sectors, not reliant on a specific dataset
    ##################################################################################
    # Tiny sectors for GOES-EAST nadir
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeseast_eqc_3km_nadir "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-EAST landocean
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeseast_eqc_3km_landocean "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-EAST longlat full disk
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeseast_longlat_50km_fulldisk "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-WEST nadir
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeswest_eqc_3km_nadir "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-WEST landocean
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeswest_eqc_3km_landocean "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-WEST longlat full disk
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goeswest_longlat_50km_fulldisk "
        "abi_netcdf "
        "Test-ABI-B14-Day-Night "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for Himawari landocean
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_himawari_eqc_3km_landocean "
        "ahi_hsd "
        "Test-AHI-B13-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_ahi/data/20200405_0000/*"
    ),
    # Tiny sectors for MeteoEU landocean, using FCI MTG-I1 dataset
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_meteoeu_eqc_3km_landocean "
        "fci_netcdf "
        "Test-FCI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_fci/data/20250623/1200/W_XX-*.nc"
    ),
    # Tiny sectors for MeteoEU longlat full disk, using SEVIRI MSG3 dataset
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_meteoeu_longlat_50km_fulldisk "
        "seviri_hrit "
        "Test-SEVIRI-B9-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_seviri/data/20250624/1200/H-000-MSG3* "
    ),
    # No MeteoIO test dataset currently
    ##################################################################################
    # ### Dataset specific test sectors, reliant on specific time of day and/or
    # ### meteorological features within the dataset.
    ##################################################################################
    # Tiny sectors for GOES-16 edge day
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes16_eqc_10km_edge_day_20200918T1950Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-16 edge night
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes16_eqc_10km_edge_night_20200918T1950Z "
        "abi_netcdf "
        "Test-ABI-B14-Night-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-16  day
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes16_eqc_3km_day_20200918T1950Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-16  night
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes16_eqc_3km_night_20200918T1950Z "
        "abi_netcdf "
        "Test-ABI-B14-Night-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-16 terminator
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes16_eqc_3km_terminator_20200918T1950Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes16_20200918_1950/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-17 edge night
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes17_eqc_10km_edge_night_20210718T0150Z "
        "abi_netcdf "
        "Test-ABI-B14-Night-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-17 edge day
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes17_eqc_10km_edge_day_20210718T0150Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-17  night
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes17_eqc_3km_night_20210718T0150Z "
        "abi_netcdf "
        "Test-ABI-B14-Night-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-17  day
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes17_eqc_3km_day_20210718T0150Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sectors for GOES-17 terminator
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_goes17_eqc_3km_terminator_20210718T0150Z "
        "abi_netcdf "
        "Test-ABI-B14-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_abi/data/goes17_20210718_0150/OR_ABI-L1b*.nc"
    ),
    # Tiny sector for Himawari terminator
    (
        "$geoips_repopath/tests/integration_tests/tiny_sectors/tiny_sectors_geostationary.sh "  # noqa: E501
        "test_himawari8_eqc_10km_terminator_20200405T0000Z "
        "ahi_hsd "
        "Test-AHI-B13-Day-Only "
        "geoips "
        "$GEOIPS_TESTDATA_DIR/test_data_ahi/data/20200405_0000/*"
    ),
]


def setup_environment():
    """
    Set up necessary environment variables for integration tests.

    Configures paths and package names for the GeoIPS core and its plugins by
    setting environment variables required for the integration tests. Assumes
    that 'GEOIPS_PACKAGES_DIR' is already set in the environment.

    Notes
    -----
    The following environment variables are set:
    - geoips_repopath
    - geoips_pkgname
    - repopath
    - pkgname
    """
    # Setup base geoips environment
    setup_geoips_environment()
    # Setup current repo's environment
    os.environ["repopath"] = os.path.realpath(
        os.path.join(os.path.dirname(__file__), "..", "..")
    )
    os.environ["pkgname"] = "geoips"


@pytest.mark.optional
@pytest.mark.sample_output
@pytest.mark.tiny_sector_evaluation
@pytest.mark.tiny_sector_global_example
@pytest.mark.parametrize("script", global_example_integ_test_calls)
def test_integ_global_example_script(base_setup: None, script: str):  # noqa: F811
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script)


@pytest.mark.optional
@pytest.mark.sample_output
@pytest.mark.tiny_sector_evaluation
@pytest.mark.tiny_sector_overlay
@pytest.mark.parametrize("script", tiny_sector_overlay_integ_test_calls)
def test_integ_tiny_sector_overlay_script(base_setup: None, script: str):  # noqa: F811
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script)


@pytest.mark.optional
@pytest.mark.full
@pytest.mark.integration
@pytest.mark.tiny_sector_evaluation
@pytest.mark.tiny_sector
@pytest.mark.parametrize("script", tiny_sector_integ_test_calls)
def test_integ_tiny_sector_script(base_setup: None, script: str):  # noqa: F811
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

# DO NOT EDIT
# managed by poetry-dynamic-versioning
__version__ = "1.17.2"
__version_tuple__ = (1, 17, 2)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Parser for dyanmic volcanoes csv files."""

import logging

import pandas as pd

LOG = logging.getLogger(__name__)

interface = "sector_metadata_generators"
family = "volc"
name = "volc_parser"


def call(trackfile_name):
    """Parse a sample volcano csv file.

    Parameters
    ----------
    trackfile_name: str, or path
        * Path to trackfile csv file with the format:
        * columns: lat lon time
        * lat/lon format: decimal
        * time format: YYYYMMDDTHHMM
    """
    csv_parser = pd.read_csv(trackfile_name)

    all_fields = []

    for i in csv_parser.iterrows():
        tmp_fields = i[1]
        fields = {}

        fields["clat"] = tmp_fields["Lat"]
        fields["clon"] = tmp_fields["Lon"]
        fields["time"] = pd.to_datetime(tmp_fields["Time"]).to_pydatetime()
        all_fields.append(fields)

    volcano_name = tmp_fields["Name"]
    if "volcano" not in volcano_name.lower():
        volcano_name = volcano_name.upper() + "_VOLCANO"
    return all_fields, volcano_name, fields["time"].strftime("%Y"), "BEST"



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

# cspell:ignore invf glle

"""Standard GeoIPS xarray dictionary based ABI NetCDF data reader."""

# Python Standard Libraries
import logging
import os
from glob import glob
from datetime import datetime, timedelta
from pathlib import Path

# Third-Party Libraries
import numpy as np
from scipy.ndimage import zoom
import xarray

from geoips.interfaces import readers
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.context_managers import import_optional_dependencies
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    check_for_partial_cache,
    check_geolocation_cache_backend,
    create_empty_partial_cache,
    get_data_cache_filename,
    get_geolocation_cache_filename,
    get_geolocation,
    rename_partial_cache,
    AutoGenError,
)


LOG = logging.getLogger(__name__)
# np.seterr(all='raise')

# Installed Libraries

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package & print to LOG.info if the import fails."""
    # If this reader is not installed on the system, don't fail alltogether,
    # just skip this import. This reader will not work if the import fails
    # and the package will have to be installed to process data of this type.
    import netCDF4 as ncdf
    import numexpr as ne
    import zarr

interface = "readers"
family = "standard"
name = "abi_netcdf"
source_names = ["abi"]

nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        f"Failed numexpr.set_num_threads in {__file__}. "
        "If numexpr is not installed and you need it, install it."
    )

DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True

# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
    "Conditional": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "No_Value": -999.6,
    "Unitialized": -9999.9,
}

DATASET_INFO = {
    "MED": ["B01", "B03", "B05"],
    "HIGH": ["B02"],
    "LOW": [
        "B04",
        "B06",
        "B07",
        "B08",
        "B09",
        "B10",
        "B11",
        "B12",
        "B13",
        "B14",
        "B15",
        "B16",
    ],
}
ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]
ALL_CHANS = {
    "LOW": [
        "B04Rad",
        "B04Ref",  # 1.37um Near-IR Cirrus
        "B06Rad",
        "B06Ref",  # 2.2um  Near-IR Cloud Particle Size
        "B07Rad",
        "B07BT",  # 3.9um  IR      Shortwave Window
        "B08Rad",
        "B08BT",  # 6.2um  IR      Upper-level tropospheric water vapor
        "B09Rad",
        "B09BT",  # 6.9um  IR      Mid-level water vapor
        "B10Rad",
        "B10BT",  # 7.3um  IR      Lower-level Water Vapor
        "B11Rad",
        "B11BT",  # 8.4um  IR      Cloud-top phase
        "B12Rad",
        "B12BT",  # 9.6um  IR      Ozone
        "B13Rad",
        "B13BT",  # 10.3um IR      Clean IR Longwave Window
        "B14Rad",
        "B14BT",  # 11.2um IR      IR Longwave window
        "B15Rad",
        "B15BT",  # 12.3um IR      Dirty Longwave Window
        "B16Rad",
        "B16BT",
    ],  # 13.3um IR      CO2 Longwave infrared
    "MED": [
        "B01Rad",
        "B01Ref",  # 0.47um Vis     Blue
        "B02Rad",
        "B02Ref",  # 0.64um Vis     Red
        "B05Rad",
        "B05Ref",
    ],  # 1.6um  Near-IR Snow/Ice
    "HIGH": ["B03Rad", "B03Ref"],  # 0.86um Near-IR Veggie
}


def metadata_to_datetime(metadata):
    """Use information from the metadata to get the image datetime."""
    times = metadata["var_info"]["time_bounds"]
    epoch = datetime(2000, 1, 1, 12, 0, 0)
    start_time = epoch + timedelta(seconds=times[0])
    end_time = epoch + timedelta(seconds=times[1])
    return start_time, end_time


def _get_files(path):
    """Get a list of file names from the input path."""
    if os.path.isfile(path):
        fnames = [path]
    elif os.path.isdir(path):
        fnames = glob(os.path.join(path, "OR_ABI*.nc"))
    elif os.path.isfile(path[0]):
        fnames = path
    else:
        raise IOError("No such file or directory: {0}".format(path))
    return fnames


def _check_file_consistency(metadata):
    """
    Check that all input metadata are from the same image time.

    Performs cheks on platform_ID, instrument_type, processing_level, and times.
    If these are all equal, returns True.
    If any differ, returns False.
    """
    # Checking file-level metadata for exact equality in the following fields
    # Was failing on processing_level due to extra NASA text sometimes
    # Also fails on time_coverage_end due to deviations on the order of 0.5 seconds
    #   Probably should fashion better checks for these two.
    # checks = ['platform_ID', 'instrument_type', 'processing_level', 'timeline_id',
    #           'scene_id', 'time_coverage_start', 'time_coverage_end']
    checks = [
        "platform_ID",
        "instrument_type",
        "timeline_id",
        "scene_id",
        "time_coverage_start",
    ]
    for name in checks:
        check_set = set(
            [metadata[fname]["file_info"][name] for fname in metadata.keys()]
        )
        if len(check_set) != 1:
            LOG.debug("Failed on {0}. Found: {1}".format(name, check_set))
            return False
    return True


def _get_file_metadata(df):
    """Gather all of the file-level metadata."""
    metadata = {}
    md_names = [
        "id",
        "dataset_name",
        "naming_authority",
        "institution",
        "project",
        "iso_series_metadata_id",
        "Conventions",
        "Metadata_Conventions",
        "keywords_vocabulary",
        "standard_name_vocabulary",
        "title",
        "summary",
        "license",
        "keywords",
        "cdm_data_type",
        "orbital_slot",
        "platform_ID",
        "instrument_type",
        "processing_level",
        "date_created",
        "production_site",
        "production_environment",
        "production_data_source",
        "timeline_id",
        "scene_id",
        "spatial_resolution",
        "time_coverage_start",
        "time_coverage_end",
    ]
    metadata_dict = {"id": "instrument_ID"}
    for name in md_names:
        try:
            if hasattr(df, name):
                metadata[name] = getattr(df, name)
            else:
                metadata[name] = getattr(df, metadata_dict[name])
        except AttributeError:
            LOG.info("Warning! File-level metadata field missing: {0}".format(name))
    return metadata


def _get_variable_metadata(df):
    """
    Gather all required variable level metadata.

    Some are skipped or gathered later as needed.
    """
    metadata = {}
    # Note: We have skipped DQF, Rad, band_wavelength_star_look, num_star_looks,
    #       star_id, t, t_star_look, x_image, x_image_bounds, y_image, y_image_bounds,
    #       geospatial_lat_lon_extent, goes_imager_projection
    md_names = [
        "band_id",
        "band_wavelength",
        "earth_sun_distance_anomaly_in_AU",
        "esun",
        "kappa0",
        "max_radiance_value_of_valid_pixels",
        "min_radiance_value_of_valid_pixels",
        "missing_pixel_count",
        "nominal_satellite_height",
        "nominal_satellite_subpoint_lat",
        "nominal_satellite_subpoint_lon",
        "percent_uncorrectable_L0_errors",
        "planck_bc1",
        "planck_bc2",
        "planck_fk1",
        "planck_fk2",
        "processing_parm_version_container",
        "saturated_pixel_count",
        "std_dev_radiance_value_of_valid_pixels",
        "time_bounds",
        "undersaturated_pixel_count",
        "valid_pixel_count",
        "x",
        "y",
        "yaw_flip_flag",
    ]
    for name in md_names:
        try:
            metadata[name] = df.variables[name][...]
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
            if name in ["x", "y"]:
                metadata[f"{name}_add_offset"] = df.variables[name].add_offset
                metadata[f"{name}_scale_factor"] = df.variables[name].scale_factor
        except KeyError:
            LOG.info("Warning! Variable-level metadata field missing: {0}".format(name))
    metadata["num_lines"] = metadata["y"].size
    metadata["num_samples"] = metadata["x"].size
    return metadata


def _get_lat_lon_extent_metadata(df):
    """Get lat lon extent metadata."""
    glle = df.variables["geospatial_lat_lon_extent"]
    metadata = {}
    md_names = [
        "geospatial_eastbound_longitude",
        "geospatial_lat_center",
        "geospatial_lat_nadir",
        "geospatial_lon_center",
        "geospatial_lon_nadir",
        "geospatial_northbound_latitude",
        "geospatial_southbound_latitude",
        "geospatial_westbound_longitude",
    ]
    for name in md_names:
        try:
            metadata[name] = getattr(glle, name)
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
        except AttributeError:
            LOG.info("Warning! Lat lon extent metadata field missing: {0}".format(name))
    return metadata


def _get_imager_projection(df):
    """Get imager projection."""
    gip = df.variables["goes_imager_projection"]
    metadata = {}
    md_names = [
        "inverse_flattening",
        "latitude_of_projection_origin",
        "longitude_of_projection_origin",
        "perspective_point_height",
        "semi_major_axis",
        "semi_minor_axis",
    ]
    for name in md_names:
        try:
            metadata[name] = getattr(gip, name)
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
        except AttributeError:
            LOG.info("Warning! Lat lon extent metadata field missing: {0}".format(name))
    metadata["grid_mapping"] = getattr(gip, "grid_mapping_name")
    return metadata


def _get_metadata(df, fname, **kwargs):
    """
    Gather metadata for the data file and return as a dictionary.

    Note: We are gathering all of the available metadata in case it is needed at
    some point.
    """
    metadata = {}
    # Gather all file-level metadata
    metadata["file_info"] = _get_file_metadata(df)
    # Gather all variable-level metadata
    metadata["var_info"] = _get_variable_metadata(df)
    # Gather lat lon extent info
    metadata["ll_info"] = _get_lat_lon_extent_metadata(df)
    # Gather projection info
    metadata["projection"] = _get_imager_projection(df)
    # Gather some useful info to the top level
    try:
        metadata["path"] = df.filepath()
    except ValueError:
        # Without cython installed, df.filepath() does not work
        metadata["path"] = fname
    metadata["satellite"] = metadata["file_info"]["platform_ID"]
    metadata["sensor"] = "ABI"
    metadata["num_lines"] = metadata["var_info"]["y"].size
    metadata["num_samples"] = metadata["var_info"]["x"].size
    return metadata


def get_latitude_longitude(
    metadata,
    BADVALS,
    sect=None,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    resource_tracker=None,
):
    """
    Get latitudes and longitudes.

    This routine accepts a dictionary containing metadata as read from a NCDF4
    format file, and returns latitudes and longitudes for a full disk.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    fname = get_geolocation_cache_filename(
        "GEOLL",
        metadata,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
    )
    if resource_tracker is not None:
        key = Path(fname).name
        if sect:
            key += f"_{sect.area_id}"
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)
    if not Path(fname).exists():
        if sect is not None and DONT_AUTOGEN_GEOLOCATION and "tc2019" not in sect.name:
            msg = (
                f"GETGEO Requested NO AUTOGEN GEOLOCATION. "
                f"Could not create latlonfile for ad {metadata['scene']}: {fname}"
            )
            LOG.error(msg)
            raise AutoGenError(msg)

        LOG.debug("Calculating latitudes and longitudes.")

        r2d = 180.0 / np.pi  # NOQA

        lambda0 = np.radians(metadata["lon0"])  # NOQA
        Re = metadata["Re"]
        # invf = metadata['invf']
        Rp = metadata["Rp"]
        # e = np.sqrt((1 / invf) * (2 - 1 / invf))
        H = metadata["H_m"]
        c = H**2 - Re**2  # NOQA

        # Python 3 netcdf reads create a masked array, while Python 2 netcdf reads
        # create ndarray. These should NOT be masked, so if we have a masked array,
        # fill it.
        if isinstance(metadata["x"], np.ma.core.MaskedArray):
            x = np.float64(metadata["x"].filled())
        else:
            x = np.float64(metadata["x"])
        if isinstance(metadata["y"], np.ma.core.MaskedArray):
            y = np.float64(metadata["y"].filled())
        else:
            y = np.float64(metadata["y"])

        LOG.info("      Making {0} by {1} grid.".format(x.size, y.size))
        # Need to transpose the latline, then repeat lonsize times
        yT = y[np.newaxis].T
        y = np.hstack([yT for num in range(x.size)])
        # Repeat lonline latsize times
        x = np.vstack([x for num in range(yT.size)])

        # Note: In this next section, we will be reusing memory space as much as
        #       possible. To make this as transparent as possible, we will do all
        #       variable assignment first, then fill them
        # This method requires that all lines remain in the SAME ORDER or things
        # will go very badly
        cosx = np.empty_like(x)
        cosy = np.empty_like(x)
        a = np.empty_like(x)
        b = np.empty_like(x)
        sinx = x  # X is not needed after the line that defines sinx
        siny = y  # Y is not needed after the line that defines siny
        rs = a
        sx = b
        sy = cosy  # sinx is not needed after the line that defines sy
        sz = cosx  # cosx is not needed after the line that defines sz
        lats = rs
        lons = sz

        LOG.info("      Calculating intermediate steps")
        Rrat = Re**2 / Rp**2  # NOQA
        ne.evaluate("cos(x)", out=cosx)  # NOQA
        ne.evaluate("cos(y)", out=cosy)  # NOQA
        ne.evaluate("sin(x)", out=sinx)  # NOQA
        ne.evaluate("sin(y)", out=siny)  # NOQA
        ne.evaluate("sinx**2 + cosx**2 * (cosy**2 + siny**2 * Rrat)", out=a)  # NOQA
        ne.evaluate("-2 * H * cosx * cosy", out=b)  # NOQA
        ne.evaluate("(-b - sqrt(b**2 - (4 * a * c))) / (2 * a)", out=rs)  # NOQA
        good_mask = np.isfinite(rs)

        ne.evaluate("rs * cosx * cosy", out=sx)  # NOQA
        ne.evaluate("rs * cosx * siny", out=sz)  # NOQA
        ne.evaluate("rs * sinx", out=sy)  # NOQA

        LOG.info("Calculating Latitudes")
        ne.evaluate("r2d * arctan(Rrat * sz / sqrt((H - sx)**2 + sy**2))", out=lats)
        LOG.info("Calculating Longitudes")
        lons = ne.evaluate("r2d * (lambda0 + arctan(sy / (H - sx)))", out=lons)
        lats[~good_mask] = BADVALS["Off_Of_Disk"]
        lons[~good_mask] = BADVALS["Off_Of_Disk"]
        LOG.info("Done calculating latitudes and longitudes")

        if geolocation_cache_backend == "memmap":
            with open(fname, "w") as df:
                lats.tofile(df)
                lons.tofile(df)
        elif geolocation_cache_backend == "zarr":
            if chunk_size:
                chunks = (chunk_size, chunk_size)
            else:
                chunks = None
            LOG.info("Storing to %s (chunks=%s)", fname, chunks)
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(fname, mode="w")
            # Assume both arrays have the same shape and dtype
            kwargs = {
                "shape": lats.shape,
                "dtype": lats.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                kwargs["chunks"] = chunks
            zf.create_dataset("lats", **kwargs)
            zf.create_dataset("lons", **kwargs)
            zf["lats"][:] = lats
            zf["lons"][:] = lons
        # # Possible switch to xarray based geolocation files, but we lose memmapping.
        # ds = xarray.Dataset(
        #     {
        #         'latitude':(['x','y'],lats),
        #         'longitude':(['x','y'],lons)
        #         }
        #     )
        # ds.to_netcdf(fname)
    else:
        # Create memmap to the lat/lon file
        # Nothing will be read until explicitly requested
        # We are mapping this here so that the lats and lons are available when
        # calculating satellite angles
        if geolocation_cache_backend == "memmap":
            shape = (metadata["num_lines"], metadata["num_samples"])
            offset = 8 * metadata["num_samples"] * metadata["num_lines"]
            lats = np.memmap(fname, mode="r", dtype=np.float64, offset=0, shape=shape)
            lons = np.memmap(
                fname, mode="r", dtype=np.float64, offset=offset, shape=shape
            )
        elif geolocation_cache_backend == "zarr":
            LOG.info(
                "GETGEO zarr to {} : lat/lon file for {}".format(
                    fname, metadata["scene"]
                )
            )
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(fname, mode="r")
            lats = zf["lats"]
            lons = zf["lons"]
    # Possible switch to xarray based geolocation files, but we lose memmapping
    # saved_xarray = xarray.load_dataset(fname)
    # lons = saved_xarray['longitude'].to_masked_array()
    # lats = saved_xarray['latitude'].to_masked_array()
    if resource_tracker is not None:
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)

    return lats, lons


def _get_geolocation_metadata(metadata):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a chash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    geomet = {}
    # G16 -> goes-16
    geomet["platform_name"] = metadata["file_info"]["platform_ID"].replace("G", "goes-")
    geomet["Re"] = metadata["projection"]["semi_major_axis"]
    geomet["Rp"] = metadata["projection"]["semi_minor_axis"]
    geomet["invf"] = metadata["projection"]["inverse_flattening"]
    geomet["e"] = 0.0818191910435
    geomet["pphgt"] = metadata["projection"]["perspective_point_height"]
    geomet["H_m"] = geomet["Re"] + geomet["pphgt"]
    geomet["lat0"] = metadata["projection"]["latitude_of_projection_origin"]
    geomet["lon0"] = metadata["projection"]["longitude_of_projection_origin"]
    geomet["scene"] = metadata["file_info"]["scene_id"]
    # Just getting the nadir resolution in kilometers.  Must extract from a string.
    geomet["res_km"] = float(
        metadata["file_info"]["spatial_resolution"].split()[0][0:-2]
    )
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    geomet["num_lines"] = metadata["var_info"]["num_lines"]
    geomet["num_samples"] = metadata["var_info"]["num_samples"]
    geomet["x"] = metadata["var_info"]["x"]
    geomet["y"] = metadata["var_info"]["y"]
    return geomet


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    geolocation_cache_backend=gpaths["GEOIPS_GEOLOCATION_CACHE_BACKEND"],
    cache_chunk_size=None,
    cache_data=False,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    roi=None,
    cache_timeout_seconds=30,
    satellite_zenith_angle_cutoff=75,
):
    """
    Read ABI NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Default=None, i.e., if not defined in the yaml file.
        * Defined in yaml file where variables and tuning parameters are set.
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to 75 for now to be consistent with previous hard coded
          value
        * Later will update to None (will have to update test scripts or
          outputs accordingly)

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
        geolocation_cache_backend=geolocation_cache_backend,
        cache_chunk_size=cache_chunk_size,
        cache_data=cache_data,
        cache_solar_angles=cache_solar_angles,
        geolocation_only=geolocation_only,
        resource_tracker=resource_tracker,
        roi=roi,
        cache_timeout_seconds=cache_timeout_seconds,
        satellite_zenith_angle_cutoff=satellite_zenith_angle_cutoff,
    )


def call_single_time(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    geolocation_cache_backend="memmap",
    cache_chunk_size=None,
    cache_data=False,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    roi=None,
    cache_timeout_seconds=30,
    satellite_zenith_angle_cutoff=None,
):
    """
    Read ABI NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    geolocation_cache_backend : str
        * Specify to use either memmap or zarray to store pre-calculated geolocation
          data.
    cache_chunk_size : int
        * Specify chunck size if using zarray to store pre-calculated geolocation data.
    resource_tracker: geoips.utils.memusg.PidLog object
        * Track resource usage using the PidLog class object from geoips.utils.memusg.
        * The PidLog.track_resource_usage method allows us to snapshot the memory usage
          for the PID associated with the geoips call. The time and stats of the
          snapshot are recorded, and can be accessed using the
          PidLog.checkpoint_usage_stats method.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Passed in from Call function.
        * Default=None, i.e., if not defined in the yaml file.
        * Defined in yaml file where variables and tuning parameters are set.
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to 75 for now to be consistent with previous hard coded
          value
        * Later will update to None (will have to update test scripts or
          outputs accordingly)

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    gvars = {}
    datavars = {}
    standard_metadata = {}
    adname = "undefined"
    if area_def and self_register:
        raise ValueError("area_def and self_register are mutually exclusive keywords")
    elif area_def:
        adname = area_def.area_id

    # Test inputs
    if area_def and self_register:
        raise ValueError(
            "sector_definition and self_register are mutually exclusive keywords"
        )

    # Get metadata for all input data files
    # Check to be sure that all input files are form the same image time
    all_metadata = {}
    for fname in fnames:
        if chans:
            gotone = False
            for chan in chans:
                currchan = (
                    chan.replace("BT", "")
                    .replace("Ref", "")
                    .replace("Rad", "")
                    .replace("B", "C")
                )
                if currchan in fname:
                    gotone = True
            if not gotone:
                LOG.info(
                    "SKIPPING file %s, not needed from channel list %s", fname, chans
                )
                continue
        try:
            with ncdf.Dataset(str(fname), "r") as df:
                all_metadata[fname] = _get_metadata(df, fname)
        except IOError as resp:
            LOG.exception("BAD FILE %s skipping", resp)
            continue
        if metadata_only:
            LOG.info("Only need metadata from first file, returning")
            break

    if not _check_file_consistency(all_metadata):
        raise ValueError("Input files inconsistent.")

    # Now put together a dict that shows what we actually got
    # This is largely to help us read in order and only create arrays of the minimum
    # required size.
    # Dict structure is channel{metadata}
    file_info = {}

    xarray_obj = xarray.Dataset()
    for md in all_metadata.values():
        ch = "B{0:02d}".format(int(md["var_info"]["band_id"]))
        if ch not in file_info:
            file_info[ch] = md

    xarray_obj.attrs["file_metadata"] = file_info.copy()

    # Most of the metadata are the same between files.
    # From here on we will just rely on the metadata from a single data file
    # for each resolution
    res_md = {}
    for res in ["LOW", "MED", "HIGH"]:
        # Find a file for this resolution: Any one will do
        res_chans = list(set(DATASET_INFO[res]).intersection(file_info.keys()))
        if res_chans:
            res_md[res] = file_info[res_chans[0]]

    # If we plan to self register, make sure we requested a resolution that we actually
    # plan to read.
    #
    # This could be problematic if we try to self-register to LOW when only
    # reading MED or something
    if self_register and self_register not in res_md:
        raise ValueError(
            "Resolution requested for self registration has not been read."
        )

    # Gather metadata
    # Assume the same for all resolutions.  Not true, but close enough.
    highest_md = res_md[list(res_md.keys())[0]]
    sdt, edt = metadata_to_datetime(highest_md)
    xarray_obj.attrs["start_datetime"] = sdt
    xarray_obj.attrs["end_datetime"] = edt
    xarray_obj.attrs["source_name"] = "abi"
    xarray_obj.attrs["data_provider"] = "noaa"
    xarray_obj.attrs["source_file_names"] = [
        os.path.basename(fname) for fname in fnames
    ]

    # additional metadata
    xarray_obj.attrs["latitude_of_projection_origin"] = all_metadata[fname][
        "projection"
    ]["latitude_of_projection_origin"]
    xarray_obj.attrs["longitude_of_projection_origin"] = all_metadata[fname][
        "projection"
    ]["longitude_of_projection_origin"]
    xarray_obj.attrs["perspective_point_height"] = all_metadata[fname]["projection"][
        "perspective_point_height"
    ]

    # G16 -> goes-16
    xarray_obj.attrs["platform_name"] = highest_md["file_info"]["platform_ID"].replace(
        "G", "goes-"
    )
    xarray_obj.attrs["area_definition"] = area_def

    # Get appropriate sector name
    if area_def:
        xarray_obj.attrs["area_id"] = area_def.area_id
        xarray_obj.attrs["area_definition"] = area_def
    else:
        if self_register and self_register not in DATASET_INFO:
            raise ValueError(
                f"Unrecognized resolution name requested for self "
                f"registration: {self_register}"
            )
        scene_id = highest_md["file_info"]["dataset_name"].split("_")[1].split("-")[2]
        if scene_id == "RadF":
            xarray_obj.attrs["area_id"] = "Full-Disk"
        elif scene_id == "RadM1":
            xarray_obj.attrs["area_id"] = "Mesoscale-1"
        elif scene_id == "RadM2":
            xarray_obj.attrs["area_id"] = "Mesoscale-2"
        elif scene_id == "RadC":
            xarray_obj.attrs["area_id"] = "CONUS"
        xarray_obj.attrs["area_definition"] = None
    adname = xarray_obj.attrs["area_id"]

    if metadata_only:
        LOG.info("Only need metadata from first file, returning")
        return {"METADATA": xarray_obj}

    # Create list of all possible channels for the case where no channels were requested
    all_chans_list = []
    for chl in ALL_CHANS.values():
        all_chans_list += chl

    # If specific channels were requested, check them against the input data
    # If specific channels were requested, but no files exist for one of the
    # channels, then error
    if chans:
        for chan in chans:
            if chan in ALL_GEO_VARS:
                continue
            if chan not in all_chans_list:
                raise ValueError("Requested channel {0} not recognized.".format(chan))
            if chan[0:3] not in file_info.keys():
                continue
                # raise ValueError(
                #     'Requested channel {0} not found in input data.'.format(chan)
                # )

    # If no specific channels were requested, get everything
    if not chans:
        chans = all_chans_list

    # Creates dict whose keys are band numbers in the form B## and whose values are
    # lists containing the data types(s) requested for the band (e.g. Rad, Ref, BT).
    chan_info = {}
    for ch in chans:
        if ch in ALL_GEO_VARS:
            continue
        chn = ch[0:3]
        typ = ch[3:]
        if chn not in chan_info:
            chan_info[chn] = []
        chan_info[chn].append(typ)

    # Gather geolocation data
    # Assume datetime the same for all resolutions.  Not true, but close enough.
    # This save us from having very slightly different solar angles for each channel.
    # Loop over resolutions and get metadata as needed
    if self_register:
        LOG.info("")
        LOG.info(
            "Getting geolocation information for resolution {} for {}.".format(
                self_register, adname
            )
        )
        # Get just the metadata we need
        standard_metadata[adname] = _get_geolocation_metadata(res_md[self_register])
        fldk_lats, fldk_lons = get_latitude_longitude(
            standard_metadata[adname],
            BADVALS,
            area_def,
            geolocation_cache_backend=geolocation_cache_backend,
            chunk_size=cache_chunk_size,
            resource_tracker=resource_tracker,
        )
        gvars[adname] = get_geolocation(
            sdt,
            standard_metadata[adname],
            fldk_lats,
            fldk_lons,
            BADVALS,
            area_def,
            geolocation_cache_backend=geolocation_cache_backend,
            cache_solar_angles=cache_solar_angles,
            scan_datetime=sdt,
            resource_tracker=resource_tracker,
        )
        if not gvars[adname]:
            LOG.error(
                f"GEOLOCATION FAILED for adname {adname} DONT_AUTOGEN_GEOLOCATION "
                f"is: {DONT_AUTOGEN_GEOLOCATION}"
            )
            gvars[adname] = {}
    else:
        for res in ["LOW", "MED", "HIGH"]:
            try:
                res_md[res]
            except KeyError:
                continue
            try:
                # Get just the metadata we need
                standard_metadata[res] = _get_geolocation_metadata(res_md[res])
                fldk_lats, fldk_lons = get_latitude_longitude(
                    standard_metadata[res],
                    BADVALS,
                    area_def,
                    geolocation_cache_backend=geolocation_cache_backend,
                    chunk_size=cache_chunk_size,
                    resource_tracker=resource_tracker,
                )
                gvars[res] = get_geolocation(
                    sdt,
                    standard_metadata[res],
                    fldk_lats,
                    fldk_lons,
                    BADVALS,
                    area_def,
                    geolocation_cache_backend=geolocation_cache_backend,
                    chunk_size=cache_chunk_size,
                    cache_solar_angles=cache_solar_angles,
                    scan_datetime=sdt,
                    resource_tracker=resource_tracker,
                )
            except ValueError as resp:
                LOG.error(
                    "{} GEOLOCATION FAILED FOR {} resolution {}. Skipping.".format(
                        resp, adname, res
                    )
                )
            # Not sure what this does, but it fails if we only pass one resolution
            # worth of data...
            #
            # MLS I think this may be necessary to avoid catastrophic failure if
            # geolocation has not been autogenerated.  I'll see.
            # if not gvars[res]:
            #     LOG.error(
            #         f"GEOLOCATION FAILED for {adname} resolution {res} "
            #         f"DONT_AUTOGEN_GEOLOCATION is: {DONT_AUTOGEN_GEOLOCATION}"
            #     )
            #     gvars[res] = {}

    LOG.interactive("Done with geolocation for {}".format(adname))
    if geolocation_only:
        geo_xarrays = {"METADATA": xarray_obj}
        for dsname in gvars.keys():
            geo_xarrays[dsname] = xarray.Dataset()
            geo_xarrays[dsname].attrs = xarray_obj.attrs.copy()
            for varname in gvars[dsname].keys():
                geo_xarrays[dsname][varname] = xarray.DataArray(gvars[dsname][varname])
        return geo_xarrays

    LOG.info("")

    # Read the data
    # Will read all data if sector_definition is None
    # Will only read required data if an sector_definition is provided
    for chan, types in chan_info.items():
        # If we didn't pass all the channel data, skip non-existent data types
        if chan not in file_info:
            continue
        LOG.info("Reading {}".format(chan))
        if resource_tracker:
            key = f"READ CHAN: abi_netcdf_chan_{chan}_{adname}"
            resource_tracker.track_resource_usage(
                logstr="MEMUSG",
                verbose=False,
                key=key,
                show_log=False,
            )
        chan_md = file_info[chan]
        for res, res_chans in DATASET_INFO.items():
            if chan in res_chans:
                break
        if (not self_register) and (
            res not in gvars.keys() or not gvars[res] or "latitude" not in gvars[res]
        ):
            LOG.info(
                f"We don't have geolocation information for {res} for {adname} "
                f"skipping {chan}."
            )
            continue
        if not area_def:
            dsname = res
        else:
            dsname = adname

        rad = ref = bt = False
        if "Rad" in types:
            rad = True
        if "Ref" in types:
            ref = True
        if "BT" in types:
            bt = True
        cache_name_prefix = f"ABI_CHAN{chan}"
        if self_register:
            data = get_data(
                chan_md,
                gvars[adname],
                rad,
                ref,
                bt,
                cache_data=cache_data,
                cache_name_prefix=cache_name_prefix,
                scan_datetime=sdt,
                standard_metadata=standard_metadata[adname],
                cache_timeout_seconds=cache_timeout_seconds,
            )
        else:
            data = get_data(
                chan_md,
                gvars[res],
                rad,
                ref,
                bt,
                cache_data=cache_data,
                cache_name_prefix=cache_name_prefix,
                scan_datetime=sdt,
                area_def=area_def,
                standard_metadata=standard_metadata[res],
                cache_timeout_seconds=cache_timeout_seconds,
            )
        for typ, val in data.items():
            if dsname not in datavars:
                datavars[dsname] = {}
            datavars[dsname][chan + typ] = val
        if resource_tracker:
            resource_tracker.track_resource_usage(
                logstr="MEMUSG", verbose=False, key=key
            )

    # This needs to be fixed:
    #   remove any unneeded datasets from datavars and gvars
    #   also mask any values below -999.0
    if area_def:
        for res in ["LOW", "MED", "HIGH"]:
            if adname not in gvars and res in gvars and gvars[res]:
                gvars[adname] = gvars[res]
            try:
                gvars.pop(res)
            except KeyError:
                pass

    # NOTE: Moved zoom and subsample to get_data, but still need to move data
    #       to correct dataset and drop extra datasets
    if self_register:
        # Determine which resolution has geolocation
        LOG.info("Registering to {}".format(self_register))
        if self_register == "HIGH":
            datavars[adname] = datavars.pop("HIGH")
            for varname, var in datavars["LOW"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 4, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 2, order=0)
            datavars.pop("MED")

        elif self_register == "MED":
            datavars[adname] = datavars.pop("MED")
            for varname, var in datavars["LOW"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 2, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["HIGH"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::2, ::2]
            datavars.pop("HIGH")

        elif self_register == "LOW":
            datavars[adname] = datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::2, ::2]
            datavars.pop("MED")
            for varname, var in datavars["HIGH"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::4, ::4]
            datavars.pop("HIGH")

        else:
            raise ValueError("No geolocation data found.")

    # basically just reformat the all_metadata dictionary to
    # reference channel names as opposed to file names..
    # band_metadata = get_band_metadata(all_metadata)

    # Remove lines and samples arrays.  Not needed.
    for res in gvars.keys():
        try:
            LOG.info(f"Popping Lines and samples for {res}")
            gvars[res].pop("Lines")
            gvars[res].pop("Samples")
        except KeyError:
            LOG.info(f"Did not find Lines and Samples to pop for {res}")
            pass
        try:
            LOG.info(
                "Checking sat zenith angles for gvar %s, cutoff specified as %s",
                res,
                satellite_zenith_angle_cutoff,
            )
            for varname, var in gvars[res].items():
                gvars[res][varname] = np.ma.array(
                    var, mask=gvars[res]["satellite_zenith_angle"].mask
                )
                if satellite_zenith_angle_cutoff:
                    LOG.info(
                        "Masking %s greater than %s degrees sat zenith angle",
                        varname,
                        satellite_zenith_angle_cutoff,
                    )
                    gvars[res][varname] = np.ma.masked_where(
                        gvars[res]["satellite_zenith_angle"]
                        > satellite_zenith_angle_cutoff,
                        gvars[res][varname],
                    )
        except KeyError:
            LOG.info(f"Did not find satellite zenith angle to mask for {res}")
            pass
    for ds in datavars.keys():
        if not datavars[ds]:
            datavars.pop(ds)

    xarray_objs = {}
    for dsname in datavars.keys():
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()
        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])
        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])

        # if roi is not defined in yaml file, a default roi is applied
        if roi is None:
            roi = 500
            roi_factor = 5
            if hasattr(xobj, "area_definition") and xobj.area_definition is not None:
                roi = max(
                    xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y
                )
                LOG.info("Trying area_def roi %s", roi)
            for curr_res in standard_metadata.keys():
                if standard_metadata[curr_res]["res_km"] * 1000.0 * roi_factor > roi:
                    roi = standard_metadata[curr_res]["res_km"] * 1000.0 * roi_factor
                    LOG.info("Trying standard_metadata[%s] %s", curr_res, roi)

        xobj.attrs["interpolation_radius_of_influence"] = roi
        LOG.info(f"Using roi {roi}")
        xarray_objs[dsname] = xobj
        # At some point we may need to deconflict, but for now just use any of the
        # dataset attributes as the METADATA dataset
        xarray_objs["METADATA"] = xobj[[]]

    LOG.info("Done reading ABI data for %s", adname)
    LOG.info("")
    return xarray_objs


def get_band_metadata(all_metadata):
    """
    Get band metadata.

    This method basically just reformats the all_metadata
    dictionary that is set based on the metadata found
    in the netcdf object itself to reference channel
    names as opposed to filenames as the dictionary keys.
    """
    bandmetadata = {}
    for fname in all_metadata.keys():
        bandnum = all_metadata[fname]["var_info"]["band_id"][0]
        bandmetadata["B%02d" % bandnum] = {}
        bandmetadata["B%02d" % bandnum] = all_metadata[fname]
    return bandmetadata


def get_cached_radiance(
    full_disk,
    line_inds,
    sample_inds,
    md,
    gvars,
    cache_name_prefix=None,
    chunk_size=None,
    scan_datetime=None,
    area_def=None,
    standard_metadata=None,
    cache_timeout_seconds=30,
):
    """Read data for a full channel's worth of files."""
    full_disk_rad_cache = get_data_cache_filename(
        cache_name_prefix + "RAD",
        standard_metadata,
        None,
        "zarr",
        chunk_size,
        scan_datetime,
    )
    cache_partial = full_disk_rad_cache + ".partial"
    if chunk_size:
        chunks = (chunk_size, chunk_size)
    else:
        chunks = None
    if not Path(full_disk_rad_cache).exists():
        if Path(cache_partial).exists():
            check_for_partial_cache(
                full_disk_rad_cache,
                cache_partial,
                cache_timeout_seconds=cache_timeout_seconds,
            )
        else:
            create_empty_partial_cache(cache_partial, "zarr")
            LOG.info("Converting netCDF to zarray")
            LOG.info("GETDATACACHE to %s", cache_partial)
            with xarray.open_dataset(md["path"]) as ds:
                ds.to_zarr(
                    cache_partial,
                    mode="w",
                    consolidated=True,
                    encoding={
                        "Rad": {"chunks": chunks, "_FillValue": -999.0},
                        "DQF": {"chunks": chunks, "_FillValue": 255},
                    },
                )
            LOG.info("Done converting netCDF to zarray")
            rename_partial_cache(cache_partial, full_disk_rad_cache)
    # This leads to redundant I/O, but should be OK. Makes thing easier rather than
    # trying to juggle either an xarray object or zarray object.
    LOG.info("GETDATACACHE from %s", full_disk_rad_cache)
    # zarr does NOT have a close method, so you can NOT use the with context.
    zf = zarr.open(full_disk_rad_cache, mode="r")
    rad_data = zf["Rad"]
    qf = zf["DQF"]
    if not full_disk:
        sector_rad_cache = get_data_cache_filename(
            cache_name_prefix + "RAD",
            standard_metadata,
            area_def,
            "zarr",
            chunk_size,
            scan_datetime,
        )
        cache_sector_partial = sector_rad_cache + ".partial"
        if (
            not Path(sector_rad_cache).exists()
            and not Path(cache_sector_partial).exists()
        ):
            create_empty_partial_cache(cache_sector_partial, "zarr")
            rad_data = rad_data[line_inds, sample_inds]
            qf = qf[line_inds, sample_inds]
            LOG.info("GETDATACACHE to %s", cache_sector_partial)
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(cache_sector_partial, mode="w")
            # Assume both arrays have the same shape and dtype
            kwargs = {
                "shape": rad_data.shape,
                "dtype": rad_data.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                kwargs["chunks"] = chunks
            zf.create_dataset("Rad", **kwargs)
            zf.create_dataset("DQF", **kwargs)
            zf["Rad"][:] = rad_data
            zf["DQF"][:] = qf
            rename_partial_cache(cache_sector_partial, sector_rad_cache)

        else:
            check_for_partial_cache(
                full_disk_rad_cache,
                cache_sector_partial,
                cache_timeout_seconds=cache_timeout_seconds,
            )
            LOG.info("GETDATACACHE from %s", sector_rad_cache)
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(sector_rad_cache, mode="r")
            rad_data = zf["Rad"]
            qf = zf["DQF"]
    else:
        # Here we need to determine which indexes to read based on the size of the
        # input geolocation data.  We assume that the geolocation data and the
        # variable data cover the same domain, just at a different resolution.
        geoloc_shape = np.array(gvars["solar_zenith_angle"].shape, dtype=np.float64)
        data_shape = np.array(rad_data.shape, dtype=np.float64)
        # If the geolocation shape matches the data shape, just read the data
        if np.all(geoloc_shape == data_shape):
            rad_data = np.float64(rad_data)
            # qf = qf
        # Upscaling data to match geolocation shape
        elif np.all(np.maximum(geoloc_shape, data_shape) == geoloc_shape):
            # Data shape is smaller, so it is the denominator
            zoom_factor, remainder = np.divmod(geoloc_shape, data_shape)
            # Ensure that all elements of the zoom factor are whole numbers
            if np.any(remainder):
                raise ValueError(
                    "Zoom factor is not a whole number.  "
                    "This section of code needs to be reexamined."
                )
            # Ensure that all elements of the zoom factor are equal
            if not np.all(zoom_factor == zoom_factor[0]):
                raise ValueError(
                    "Zoom factor must be equal for both dimensions.  "
                    "This section of code needs to be reexamined."
                )
            # Perform the actual zoom
            zoom_factor = zoom_factor.astype(np.int)
            rad_data = zoom(np.float64(rad_data), zoom_factor[0], order=0)
            qf = zoom(qf, zoom_factor[0], order=0)
        # Downscaling data to match geolocation shape
        elif np.all(np.maximum(geoloc_shape, data_shape) == data_shape):
            # Geoloc shape is smaller, so it is the denominator
            zoom_factor, remainder = np.divmod(data_shape, geoloc_shape)
            # Ensure that all elements of the zoom factor are whole numbers
            if np.any(remainder):
                raise ValueError(
                    "Zoom factor is not a whole number.  "
                    "This section of code needs to be reexamined."
                )
            # Ensure that all elements of the zoom factor are equal
            if not np.all(zoom_factor == zoom_factor[0]):
                raise ValueError(
                    "Zoom factor must be equal for both dimensions.  "
                    "This section of code needs to be reexamined."
                )
            # Perform the actual subsampling
            LOG.info("Before zoom")
            zoom_factor = zoom_factor.astype(np.int)
            rad_data = np.float64(rad_data[:: zoom_factor[0], :: zoom_factor[0]])
            qf = qf[:: zoom_factor[0], :: zoom_factor[0]]
            LOG.info("After zoom")
        else:
            raise ValueError(
                "Zoom factor cannot be computed.  "
                "All either both dimensions of geolocation data must be "
                "larger than the data dimensions or vice versa."
            )
    return np.float64(rad_data), np.float64(qf)


def read_netcdf_radiance(full_disk, line_inds, sample_inds, md, gvars):
    """Read data for a full channel's worth of files."""
    # Read radiance data for channel
    # For some reason, netcdf4 does not like it if you pass a bunch of indexes.
    # It is very slow and creates memory errors if there are too many.
    # Have to read ALL of the data, then subset.
    # Need to find a solution for this.
    # Use with to open the data file for reading and automatically close once complete
    with ncdf.Dataset(md["path"], "r") as df:
        if not full_disk:
            rad_data = np.float64(df.variables["Rad"][...][line_inds, sample_inds])
            qf = df.variables["DQF"][...][line_inds, sample_inds]
        else:
            # Here we need to determine which indexes to read based on the size of the
            # input geolocation data.  We assume that the geolocation data and the
            # variable data cover the same domain, just at a different resolution.
            geoloc_shape = np.array(gvars["solar_zenith_angle"].shape, dtype=np.float64)
            data_shape = np.array(df.variables["Rad"].shape, dtype=np.float64)
            # If the geolocation shape matches the data shape, just read the data
            if np.all(geoloc_shape == data_shape):
                rad_data = np.float64(df.variables["Rad"][...])
                qf = df.variables["DQF"][...]
            # Upscaling data to match geolocation shape
            elif np.all(np.maximum(geoloc_shape, data_shape) == geoloc_shape):
                # Data shape is smaller, so it is the denominator
                zoom_factor, remainder = np.divmod(geoloc_shape, data_shape)
                # Ensure that all elements of the zoom factor are whole numbers
                if np.any(remainder):
                    raise ValueError(
                        "Zoom factor is not a whole number.  "
                        "This section of code needs to be reexamined."
                    )
                # Ensure that all elements of the zoom factor are equal
                if not np.all(zoom_factor == zoom_factor[0]):
                    raise ValueError(
                        "Zoom factor must be equal for both dimensions.  "
                        "This section of code needs to be reexamined."
                    )
                # Perform the actual zoom
                zoom_factor = zoom_factor.astype(np.int)
                rad_data = zoom(
                    np.float64(df.variables["Rad"][...]), zoom_factor[0], order=0
                )
                qf = zoom(df.variables["DQF"][...], zoom_factor[0], order=0)
            # Downscaling data to match geolocation shape
            elif np.all(np.maximum(geoloc_shape, data_shape) == data_shape):
                # Geoloc shape is smaller, so it is the denominator
                zoom_factor, remainder = np.divmod(data_shape, geoloc_shape)
                # Ensure that all elements of the zoom factor are whole numbers
                if np.any(remainder):
                    raise ValueError(
                        "Zoom factor is not a whole number.  "
                        "This section of code needs to be reexamined."
                    )
                # Ensure that all elements of the zoom factor are equal
                if not np.all(zoom_factor == zoom_factor[0]):
                    raise ValueError(
                        "Zoom factor must be equal for both dimensions.  "
                        "This section of code needs to be reexamined."
                    )
                # Perform the actual subsampling
                LOG.info("Before zoom")
                zoom_factor = zoom_factor.astype(np.int)

                # NOTE: Strides are broken for netCDF4 library version < 4.6.2.
                #       At present, the most recent stable release is 4.6.1.
                #       Once 4.6.2 is released, this should be retested.
                #       See https://github.com/Unidata/netcdf4-python/issues/680

                rad_data = np.float64(
                    df.variables["Rad"][...][:: zoom_factor[0], :: zoom_factor[0]]
                )
                qf = df.variables["DQF"][...][:: zoom_factor[0], :: zoom_factor[0]]
                # rad_data = np.float64(
                #     df.variables['Rad'][::zoom_factor[0], ::zoom_factor[0]]
                # )
                # qf = df.variables['DQF'][::zoom_factor[0], ::zoom_factor[0]]
                LOG.info("After zoom")
            else:
                raise ValueError(
                    "Zoom factor cannot be computed.  "
                    "All either both dimensions of geolocation data must be "
                    "larger than the data dimensions or vice versa."
                )
    return rad_data, qf


def convert_radiance_to_reflectance(
    radiance, solar_zenith_angle, calibration_metadata, bad_data_mask
):
    """Convert radiance to reflectance.

    Parameters
    ----------
    radiance : numpy.masked_array
        Observed radiances
    solar_zenith_angle : numpy.array
        Calculated solar angles for each radiance pixel
    calibration_metadata : dict
        Dictionary holding the kappa0 constant used when converting radiance to BT
    bad_data_mask : numpy.array
        Bool array where bad data are masked

    Returns
    -------
    numpy.array
        Calibrated reflectances
    """
    # Get the radiance data
    # Have to do this when using numexpr
    rad_data = radiance[~bad_data_mask]  # NOQA

    k0 = calibration_metadata["kappa0"]  # NOQA
    sun_zenith = solar_zenith_angle[~bad_data_mask]  # NOQA
    # zoom_info = (
    #     np.array(rad_data.shape, dtype=np.float) /
    #     np.array(sun_zenith.shape, dtype=np.float
    # )
    # if zoom_info[0] == zoom_info[1] and int(zoom_info[0]) == zoom_info[0]:
    #     if zoom_info[0] > 1 and int(zoom_info[0]) == zoom_info[0]:
    #         sun_zenith = zoom(sun_zenith, zoom_info[0])
    #     elif zoom_info[0] < 1 and int(zoom_info[0]**-1) == zoom_info[0]**-1:
    #         sun_zenith = sun_zenith[::zoom_info[0]**-1, :: zoom_info[0]**-1]
    #     elif zoom_info[0] == 1:
    #         pass
    #     else:
    #         ValueError('Inappropriate zoom level calculated.')
    # else:
    #     ValueError('Inappropriate zoom level calculated.')

    deg2rad = np.pi / 180.0  # NOQA
    # NOTE:: ABI docs suggest doing solar zenith angle correction here, but
    #        since some algorithms (e.g. Fire-Temperature RGB) required uncorrected
    #        data, we will not do this here, instead leaving it to the algorithm
    #        developer to handle.
    ref = ne.evaluate("k0 * rad_data")
    return ref


def get_cached_reflectance(
    radiance,
    solar_zenith_angle,
    calibration_metadata,
    bad_data_mask,
    cache_name_prefix=None,
    chunk_size=None,
    scan_datetime=None,
    area_def=None,
    standard_metadata=None,
    cache_timeout_seconds=30,
):
    """Create or load cached zarray of reflectance data.

    Loads zarray file if cache exists, otherwise calls convert_radiance_to_reflectance
    to calculate the reflectances then stores to zarray.

    Parameters
    ----------
    radiance : numpy.masked_array
        Observed radiances
    solar_zenith_angle : numpy.array
        Calculated solar angles for each radiance pixel
    calibration_metadata : dict
        Dictionary holding the kappa0 constant used when converting radiance to BT
    bad_data_mask : numpy.array
        Bool array where bad data are masked
    cache_name_prefix : str, optional
        Cache file name prefix, by default None
    chunk_size : int, optional
        Chunk size of zarray, by default None
    scan_datetime : datetime, optional
        Scan start time of observation, by default None
    area_def : pyresample.area_definition, optional
        Area definition subsector, by default None
    standard_metadata : dict, optional
        High level metadata for dataset, by default None

    Returns
    -------
    array
        Calibrated reflectances
    """
    ref_cache = get_data_cache_filename(
        cache_name_prefix + "REF",
        standard_metadata,
        area_def,
        "zarr",
        chunk_size,
        scan_datetime,
    )
    cache_partial = ref_cache + ".partial"
    if Path(ref_cache).exists() or Path(cache_partial).exists():
        check_for_partial_cache(
            ref_cache, cache_partial, cache_timeout_seconds=cache_timeout_seconds
        )
        LOG.info("GETDATACACHE from %s", ref_cache)
        # zarr does NOT have a close method, so you can NOT use the with context.
        zf = zarr.open(ref_cache, mode="r")
        ref = zf["Ref"]
    else:
        create_empty_partial_cache(cache_partial, "zarr")
        if chunk_size:
            chunks = (chunk_size, chunk_size)
        else:
            chunks = None
        LOG.info("GETDATACACHE to %s", cache_partial)
        ref = convert_radiance_to_reflectance(
            radiance, solar_zenith_angle, calibration_metadata, bad_data_mask
        )

        kwargs = {
            "shape": ref.shape,
            "dtype": ref.dtype,
        }
        # zarr does NOT have a close method, so you can NOT use the with context.
        zf = zarr.open(cache_partial, mode="w")
        # As of Python 3.11, can't pass chunks=None into create_dataset
        if chunks:
            kwargs["chunks"] = chunks
        zf.create_dataset("Ref", **kwargs)
        zf["Ref"][:] = ref
        rename_partial_cache(cache_partial, ref_cache)

    return ref


def convert_radiance_to_brightness_temperature(
    radiance, calibration_metadata, bad_data_mask
):
    """Convert radiance to brightness temperature.

    Parameters
    ----------
    radiance : numpy.masked_array
        Observed radiances
    calibration_metadata : dict
        Dictionary holding the planck_fk1, planck_fk2, planck_bc1, and planck_bc2
        constants that are used for converting radiance to BT
    bad_data_mask : numpy.array
        Bool array where bad data are masked

    Returns
    -------
    numpy.array
        Calibrated brightness temperatures
    """
    # Get the radiance data
    # Have to do this when using numexpr
    rad_data = radiance[~bad_data_mask]  # NOQA

    # Sometimes negative values occur due to oddities in calibration
    # These break the conversion to BT
    # Since these values are likely caused by very dark (cold) scenes
    #   that are below the sensor's sensitivity and we are not interested in
    #   having bad data scattered throughout the imager, we will set this
    #   to a very small value
    rad_data[rad_data <= 0] = 0.001

    fk1 = calibration_metadata["planck_fk1"]  # NOQA
    fk2 = calibration_metadata["planck_fk2"]  # NOQA
    bc1 = calibration_metadata["planck_bc1"]  # NOQA
    bc2 = calibration_metadata["planck_bc2"]  # NOQA

    bt = ne.evaluate("(fk2 / log(fk1 / rad_data + 1) - bc1) / bc2")
    return bt


def get_cached_brightness_temperature(
    radiance,
    calibration_metadata,
    bad_data_mask,
    cache_name_prefix=None,
    chunk_size=None,
    scan_datetime=None,
    area_def=None,
    standard_metadata=None,
    cache_timeout_seconds=30,
):
    """Create or load cached zarray of brightness temperature data.

    Loads zarray file if cache exists, otherwise calls
    convert_radiance_to_brightness_temperature to calculate the reflectances then stores
    to zarray.

    Parameters
    ----------
    radiance : numpy.masked_array
        Observed radiances
    calibration_metadata : dict
        Dictionary holding the planck_fk1, planck_fk2, planck_bc1, and planck_bc2
        constants that are used for converting radiance to BT
    bad_data_mask : numpy.array
        Bool array where bad data are masked
    cache_name_prefix : str, optional
        Cache file name prefix, by default None
    chunk_size : int, optional
        Chunk size of zarray, by default None
    scan_datetime : datetime, optional
        Scan start time of observation, by default None
    area_def : pyresample.area_definition, optional
        Area definition subsector, by default None
    standard_metadata : dict, optional
        High level metadata for dataset, by default None

    Returns
    -------
    array
        Calibrated brightness temperatures
    """
    bt_cache = get_data_cache_filename(
        cache_name_prefix + "BT",
        standard_metadata,
        area_def,
        "zarr",
        chunk_size,
        scan_datetime,
    )

    cache_partial = bt_cache + ".partial"
    if Path(bt_cache).exists() or Path(cache_partial).exists():
        check_for_partial_cache(
            bt_cache, cache_partial, cache_timeout_seconds=cache_timeout_seconds
        )
        LOG.info("GETDATACACHE from %s", bt_cache)
        # zarr does NOT have a close method, so you can NOT use the with context.
        zf = zarr.open(bt_cache, mode="r")
        bt = zf["BT"]
    else:
        create_empty_partial_cache(cache_partial, "zarr")
        if chunk_size:
            chunks = (chunk_size, chunk_size)
        else:
            chunks = None
        LOG.info("GETDATACACHE to %s", cache_partial)
        bt = convert_radiance_to_brightness_temperature(
            radiance, calibration_metadata, bad_data_mask
        )

        # zarr does NOT have a close method, so you can NOT use the with context.
        zf = zarr.open(cache_partial, mode="w")
        kwargs = {
            "shape": bt.shape,
            "dtype": bt.dtype,
        }
        # As of Python 3.11, can't pass chunks=None into create_dataset
        if chunks:
            kwargs["chunks"] = chunks
        zf.create_dataset("BT", **kwargs)
        zf["BT"][:] = bt
        rename_partial_cache(cache_partial, bt_cache)

    return bt


def get_data(
    md,
    gvars,
    rad=False,
    ref=False,
    bt=False,
    cache_data=False,
    cache_name_prefix=None,
    chunk_size=None,
    scan_datetime=None,
    area_def=None,
    standard_metadata=None,
    cache_timeout_seconds=30,
):
    """Read data for a full channel's worth of files."""
    # Coordinate arrays for reading
    if "Lines" in gvars and "Samples" in gvars:
        full_disk = False
        line_inds = gvars["Lines"]
        sample_inds = gvars["Samples"]
    else:
        full_disk = True
        line_inds = []
        sample_inds = []

    band_num = md["var_info"]["band_id"]

    if cache_data:
        rad_data, qf = get_cached_radiance(
            full_disk,
            line_inds.copy(),
            sample_inds.copy(),
            md,
            gvars,
            cache_name_prefix=cache_name_prefix,
            chunk_size=chunk_size,
            scan_datetime=scan_datetime,
            area_def=area_def,
            standard_metadata=standard_metadata,
            cache_timeout_seconds=cache_timeout_seconds,
        )
    else:
        rad_data, qf = read_netcdf_radiance(
            full_disk, line_inds, sample_inds, md, gvars
        )
    data = {"Rad": rad_data}
    # Originally was using -1, but this is a uint8 and set to 255 for off of disk
    data["Rad"][np.where(qf == 255)] = BADVALS["Off_Of_Disk"]
    data["Rad"][np.where(qf == 1)] = BADVALS["Conditional"]
    # This flag is ignored for now.  It masks good data that is useful for imagery.
    # I am unsure if this should be used in digital products.
    # data['Rad'][np.where(qf == 2)] = BADVALS['Out_Of_Valid_Range']
    data["Rad"][np.where(qf == 3)] = BADVALS["No_Value"]
    data["Rad"] = np.ma.masked_less_equal(data["Rad"], -999.0)

    # Get the bad data mask from radiances to reuse
    bad_data_mask = data["Rad"].mask

    # If reflectance is requested
    # Note, weird memory manipulations to save memory when radiances are not requested
    if ref:
        if band_num not in range(1, 7):
            raise ValueError(
                "Unable to calculate reflectances for band #{0}".format(band_num)
            )
        # If we don't need radiances, then reuse the memory
        if not rad:
            data["Ref"] = data.pop("Rad")
            radiance = data["Ref"]
        else:
            data["Ref"] = np.empty_like(data["Rad"])
            radiance = data["Rad"]
        if cache_data:
            ref = get_cached_reflectance(
                radiance,
                gvars["solar_zenith_angle"],
                md["var_info"],
                bad_data_mask,
                cache_name_prefix=cache_name_prefix,
                chunk_size=chunk_size,
                scan_datetime=scan_datetime,
                area_def=area_def,
                standard_metadata=standard_metadata,
                cache_timeout_seconds=cache_timeout_seconds,
            )
        else:
            ref = convert_radiance_to_reflectance(
                radiance, gvars["solar_zenith_angle"], md["var_info"], bad_data_mask
            )
        data["Ref"][~bad_data_mask] = ref

    if bt:
        if band_num not in range(7, 17):
            raise ValueError(
                "Unable to calculate brightness temperatures for band #{0}".format(
                    band_num
                )
            )
        # If we don't need radiances, then reuse the memory
        if not rad:
            data["BT"] = data.pop("Rad")
            radiance = data["BT"]
        else:
            data["BT"] = np.empty_like(data["Rad"])
            radiance = data["Rad"]

        if cache_data:
            bt = get_cached_brightness_temperature(
                radiance,
                md["var_info"],
                bad_data_mask,
                cache_name_prefix=cache_name_prefix,
                chunk_size=chunk_size,
                scan_datetime=scan_datetime,
                area_def=area_def,
                standard_metadata=standard_metadata,
                cache_timeout_seconds=cache_timeout_seconds,
            )
        else:
            bt = convert_radiance_to_brightness_temperature(
                radiance, md["var_info"], bad_data_mask
            )

        data["BT"][~bad_data_mask] = bt

    # latitude is sometimes fully specified from area_def... So does not
    # relate to actual masked data..
    # This should only occur when area_def was provided and breaks when it wasn't.
    if "Lines" in gvars:
        usemask = np.ma.masked_less(gvars["Lines"], -990).mask
    else:
        usemask = gvars["latitude"].mask

    for varname in gvars.keys():
        if varname in ["latitude", "longitude"]:
            continue
        if hasattr(gvars[varname], "mask"):
            gvars[varname].mask = usemask
    for varname in data.keys():
        if hasattr(data[varname], "mask"):
            data[varname].mask = usemask
    return data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""geoips.commandline.ancillary_info init file."""
import os
from geoips.utils.cache_files import get_cached_json


# This will collect ancillary information contained in YAML files, but will convert them
# to a cached JSON file first. This is done because reading YAML files is slow.
alias_mapping = get_cached_json(f"{os.path.dirname(__file__)}/alias_mapping.yaml")
cmd_instructions = get_cached_json(f"{os.path.dirname(__file__)}/cmd_instructions.yaml")



"""Pytest configuration for test data validation.

Provides command-line options and fixtures to control test behavior
when required test data files are missing.
"""

import pytest


def pytest_addoption(parser):
    """Add custom command-line options to pytest.

    Parameters
    ----------
    parser : pytest.Parser
        Pytest command-line argument parser.

    Notes
    -----
    Adds --no-fail-on-missing-data flag to skip tests instead of
    failing when test data is unavailable.
    """
    parser.addoption(
        "--no-fail-on-missing-data",
        action="store_true",
        default=False,
        help="Fail tests if test data is missing",
    )


@pytest.fixture
def fail_on_missing_data(request):
    """Determine whether to fail tests on missing data.

    Parameters
    ----------
    request : pytest.FixtureRequest
        Pytest fixture request object.

    Returns
    -------
    bool
        True if tests should fail on missing data, False to skip instead.
    """
    return not request.config.getoption("--no-fail-on-missing-data")



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Generic testing utilities used for pydantic unit testing."""

import logging
from importlib import import_module, resources
from pathlib import Path
import pytest
from copy import deepcopy
from typing import Any, Tuple, Type
import warnings
import yaml

from pydantic import BaseModel, Field, model_validator, ValidationError

from geoips import interfaces
from geoips import pydantic_models as geoips_models


LOG = logging.getLogger(__name__)


class PathDict(dict):
    """Custom dictionary class used to access GeoIPS plugins using paths.

    Where a 'path' is a key which can include slashes ('/') for easy access to the
    values contained in nested dictionaries.

    I.e. some_dict["key1/key2/key3"] = some_val

    This is essentially the same thing as a dict object besides the __setitem__ method.
    Not inheriting from that class to avoid unexpected consequences.
    """

    def __setitem__(self, key, value):
        """Set an item under key with value 'value'.

        Custom implementation of dict's __setitem__, which makes use of paths to
        set an item which may be multiple levels deep.

        Parameters
        ----------
        key: str
            - The path to the attribute to set
        value: any
            - The value to set the attribute to.
        """
        keys = key.split("/")
        if len(keys) > 1:
            current = self
            for subkey in keys[:-1]:
                current = current.setdefault(subkey, {})
            dict.__setitem__(current, keys[-1], value)
        else:
            dict.__setitem__(self, key, value)


class TestCaseModel(BaseModel):
    """YAML-based test case input model validation.

    This pydantic model validates the structure of test case inputs defined in YAML. It
    ensures that required keys and value types are correctly specified for each test
    case.
    """

    test_case_id: str = Field(
        ..., description="Unique identifier for the individual test case."
    )
    description: str = Field(
        None, description="Short description of the test case and its purpose."
    )
    key: str = Field(..., description="Path to the attribute being mutated.")
    val: Any = Field(..., description="The value to set for the given key.")
    cls: object = Field(..., description="The name of the model expected to fail.")
    err_str: str = Field(
        ..., description="Expected error message fragment for the ValidationError."
    )

    @model_validator(mode="before")
    @classmethod
    def _warn_if_missing_description(cls, values):
        if "description" not in values:
            warnings.warn(
                f"Test case '{values.get('test_case_id', '[unknown]')}' is missing a"
                f"description. This warning will become an error in GeoIPS 1.18.",
                UserWarning,
            )
        return values


def load_test_cases(interface_name: str, test_type: str) -> dict:
    """Load a set of test cases used to validate pydantic model(s).

    This can either be a top level model, such as SectorPluginModel, or a component
    of that model.

    Parameters
    ----------
    interface_name: str
        - The name of the interface that's going to be tested.
    test_type: str
        - The category of the test cases to load - either "bad" or "neutral".

    Returns
    -------
    test_cases: dict
        - The dictionary of test cases used to validate your model.
    """
    if test_type not in ("bad", "neutral"):
        raise ValueError(f"Unsupported test type: {test_type}")

    fname = f"test_cases_{test_type}.yaml"
    base_dir = Path(__file__).parent
    fpath = base_dir / interface_name / fname

    if not fpath.exists():
        raise FileNotFoundError(
            f"Error: No test cases file could be found. Expected {fpath} but it "
            "does not exist. Please create this file and rerun your tests."
        )

    with fpath.open("r") as fo:
        raw_test_cases = yaml.safe_load(fo)

    validated_test_cases = {}
    for test_case_id, raw_test_case in raw_test_cases.items():
        try:
            raw_test_case["test_case_id"] = test_case_id
            validated_test_case = TestCaseModel(**raw_test_case)
            validated_test_cases[test_case_id] = validated_test_case
        except ValidationError as e:
            raise RuntimeError(f"Invalid test case '{test_case_id}': {e}")

    return validated_test_cases


def load_geoips_yaml_plugin(interface_name: str, plugin_name: str) -> dict:
    """Load a GeoIPS YAML plugin via yaml.safe_load, not interface.get_plugin.

    This will be used until we convert all of our schema to pydantic. If we load a
    plugin via <interface>.get_plugin, this sometimes will cause errors due to different
    validation protocols.

    TODO: Turn on a flag that just disables validation in 'get_plugin'

    Parameters
    ----------
    interface_name: str
        - The name of the GeoIPS plugin's interface.
    plugin_name: str
        - The name of the plugin of type 'interface_name'.

    Returns
    -------
    yam: dict
        - A dictionary object representing the yaml plugin requested.
    """
    yam_ints = interfaces.list_available_interfaces()["yaml_based"]
    if interface_name not in yam_ints:
        raise KeyError(
            f"'{interface_name}' is not a valid yaml interface. Please select an "
            f"interface out of any of the following {yam_ints}."
        )
    interface = getattr(interfaces, interface_name)
    registry = interface.plugin_registry.registered_plugins["yaml_based"][
        interface_name
    ]
    if isinstance(plugin_name, tuple):
        # Likely a product. Do a nested search for the plugin.
        entry = registry[plugin_name[0]][plugin_name[1]]
    else:
        entry = registry[plugin_name]

    relpath = entry["relpath"]
    print("relpaht \t", relpath)
    abspath = str(resources.files("geoips") / relpath)
    package = "geoips"

    with open(abspath, "r") as fo:
        yam = yaml.safe_load(fo)
    # Append these attributes to the plugin for further validation. 'get_plugin'
    # already does this, but since we're loading in a different manner, we do this
    # manually for the time being.
    yam["abspath"] = abspath
    yam["relpath"] = relpath
    yam["package"] = package

    return yam


def retrieve_model(plugin):
    """Retrieve the associated pydantic plugin model for the plugin provided.

    Parameters
    ----------
    plugin: dict
        - A dictionary representing a GeoIPS yaml plugin.

    Returns
    -------
    model: geoips.pydantic.{interface} PluginModel
        - The associated plugin model used to validate this plugin.
    """
    interface = plugin["interface"]
    # upcoming PR: https://github.com/NRLMMD-GEOIPS/geoips/issues/1125
    # module = geoips_models._modules[f"geoips.pydantic_models.v1.{interface}"]
    module = import_module(f"geoips.pydantic_models.v1.{interface}")
    if "_" in interface:
        int_split = interface.split("_")
        interface = f"{int_split[0].title()}{int_split[1].title()}"
    else:
        interface = interface.title()
    # Remove leading and trailing whitespace and remove the suffix 's' if it exists
    # So, if we had 'sectors', it would become 'sector'
    interface = interface.rstrip().removesuffix("s")
    model = getattr(module, f"{interface}PluginModel")

    return model


def _validate_test_tup_keys(test_tup: TestCaseModel) -> Tuple[str, Any, str, str]:
    """Ensure test_tup contains either 'err_str' or 'warn_match', and extract values.

    Parameters
    ----------
    test_tup:
        - A dictionary containing the following keys:
            - 'key' (str): Path to the attribute being mutated.
            - 'val' (any): The value to set for the given key.
            - 'cls' (str): The name of the model expected to fail.
            - 'err_str' (str): Expected error message fragment for the
                ValidationError.

    Returns
    -------
    tuple
        (key, value, cls, err_str / warn_match)

    Raises
    ------
    ValueError
        If neither or both 'err_str' and 'warn_match' are provided.
    """
    key = test_tup.key
    val = test_tup.val
    failing_model = test_tup.cls
    err_str = test_tup.err_str

    return key, val, failing_model, err_str


def _resolve_model_class(failing_model):
    """Resolve the actual Pydantic model class by name."""
    for mod_name, mod in geoips_models.v1._modules.items():
        if failing_model in geoips_models.v1._classes.get(mod_name, {}):
            return getattr(mod, failing_model)
        if hasattr(mod, failing_model):
            # This behavior occurs for the 'ColorType' attribute, which is
            # a type instance but not actually a pydantic class.
            return None
    return None


def _attempt_to_associate_model_with_error(
    failing_model: str, errors: list[dict]
) -> dict:
    """Attempt to associate the correct error with the model that should be failing.

    This occurs when more than one error was raised for a given test. Assuming that
    two fields were not incorrect, only one, then this means that a field is one of a
    Union of models and/or other fields. Attempt to locate the correct error based on
    the model provided.

    Parameters
    ----------
    failing_model: str
        - The name (case sensitive) of the model that we expect to raise this error
    errors: pydantic.ValidationError.errors() -- list
        - A list of errors containing information on why a field failed to validate.

    """
    for error in reversed(errors):
        if failing_model in error.get("loc", ()):
            return error

    # If no errors could be associated with the failing model, just default
    # to the last error reported
    return errors[-1]


def validate_base_plugin(base_plugin: dict, plugin_model: Type):
    """Run a base test to assert that a well-formatted plugin is valid.

    Parameters
    ----------
    base_plugin: dict
        - A dictionary representing a valid plugin.
    plugin_model: instance or child of geoips.pydantic.bases.PluginModel
        - The pydantic-based model used to validate this plugin.
    """
    plugin_model(**base_plugin)


def validate_neutral_plugin(
    base_plugin: dict, test_tup: Tuple[str, Any, str, str], plugin_model: type
):
    """Run a neutral test to check for expected FutureWarnings in a plugin.

    Parameters
    ----------
    base_plugin: dict
        - A dictionary representing a plugin that is valid.
    test_tup:
        - A tuple formatted (key, value, class, err_str), formatted (str, any, str, str)
          used to run and validate tests.
    plugin_model: instance or child of geoips.pydantic.bases.PluginModel
        - The pydantic-based model used to validate this plugin.
    """
    key, val, failing_model, err_str = _validate_test_tup_keys(test_tup)

    neutral_plugin = deepcopy(base_plugin)
    neutral_plugin[key] = val

    if err_str:
        with pytest.warns(FutureWarning, match=err_str):
            plugin_model(**neutral_plugin)
        return


def validate_bad_plugin(
    base_plugin: dict, test_tup: Tuple[str, Any, str, str], plugin_model: Type
):
    """Perform validation on any GeoIPS plugin, ensuring correct ValidationErrors occur.

    Current supported plugins include readers and sectors. More to come in the future,
    and this function might change because of that. This is a beta method for the time
    being.

    Parameters
    ----------
    base_plugin: dict
        - A dictionary representing a plugin that is valid.
    test_tup:
        - A tuple formatted (key, value, class, err_str), formatted (str, any, str, str)
          used to run and validate tests.
    plugin_model: instance or child of geoips.pydantic.bases.PluginModel
        - The pydantic-based model used to validate this plugin.
    """
    key, val, failing_model, err_str = _validate_test_tup_keys(test_tup)

    bad_plugin = deepcopy(base_plugin)
    bad_plugin[key] = val

    # ValidationErrors won't occur for these fields at the moment.
    # Skip the testing until the validators are implemented.
    validation_to_be_implemented = ["abspath", "relpath", "package"]
    if key in validation_to_be_implemented:
        return

    try:
        plugin_model(**bad_plugin)
    except ValidationError as e:
        # The code below assumes that your test only raised one error. That's how
        # we've structured testing for the time being. In the case that one or more
        # errors are reported, we default to the last error of the failing model
        # reported, or, if no failing model could be associated with this error, we
        # just default to the last error reported.
        errors = e.errors()
        if len(e.errors()) > 1:
            val_err = _attempt_to_associate_model_with_error(failing_model, errors)
        else:
            val_err = errors[0]
        # In Pydantic ValidationError, the last element of 'loc' tuple identifies
        # the failing attribute
        bad_field = val_err["loc"][-1]
        err_msg = val_err["msg"]

        model_class = _resolve_model_class(failing_model)

        if model_class:
            assert (
                bad_field in model_class.model_fields
            ), f"Field '{bad_field}' not found in model '{failing_model}'"
        if err_str:
            assert (
                err_str in err_msg or err_str == err_msg
            ), f"Expected error message to match: \n {err_str} \n Got:\n{err_msg}"
    else:
        assert False, f"Expected ValidationError for key '{key}', but none was raised."



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI help messages."""

import argparse

# Odd import done to prevent re-doc of imported class
# which causes a sphinx error :')
from geoips.commandline.geoips_command import AlphabeticalHelpFormatter


class TestAlphabeticalHelpFormatter:
    """
    Test suite for AlphabeticalHelpFormatter.

    Ensures that the custom help formatter correctly sorts
    command-line arguments in alphabetical order.

    Tests cover:
    - Mixed argument types (options and positional arguments)
    - Arguments with multiple aliases
    - Complex option string configurations
    """

    def test_sorting_with_mixed_arguments(self):
        """
        Test sorting of arguments with mixed types and flags.

        Verifies alphabetical sorting of:
        - Options with short and long flags
        - Positional arguments
        - Arguments with multiple aliases
        """
        parser = argparse.ArgumentParser(formatter_class=AlphabeticalHelpFormatter)

        # Add arguments in non-alphabetical order with various types
        parser.add_argument("-z", "--zeta", help="Zeta option")
        parser.add_argument("gamma", help="Gamma positional argument")
        parser.add_argument("-a", "--alpha", help="Alpha option")
        parser.add_argument("beta", help="Beta positional argument")
        parser.add_argument("-m", "--middle", help="Middle option")

        # Capture help output
        help_output = parser.format_help()
        help_output = help_output[help_output.index("\n") :]  # skip first line

        # Verify alphabetical ordering
        assert help_output.index("--alpha") < help_output.index("--middle")
        assert help_output.index("--middle") < help_output.index("--zeta")
        assert help_output.index("beta") < help_output.index("gamma")

    def test_sorting_with_complex_option_strings(self):
        """
        Test sorting of arguments with multiple option strings and aliases.

        Ensures alphabetical ordering based on first option string.
        """
        parser = argparse.ArgumentParser(formatter_class=AlphabeticalHelpFormatter)

        # Add arguments with multiple option strings and aliases
        parser.add_argument("-z", "--zeta", "--zebra", help="Zeta option")
        parser.add_argument("-a", "--alpha", "--apple", help="Alpha option")
        parser.add_argument("-m", "--middle", "--monkey", help="Middle option")

        # Capture help output
        help_output = parser.format_help()
        help_output = help_output[help_output.index("\n") :]  # skip first line

        # Verify alphabetical ordering based on first option string
        assert help_output.index("--alpha") < help_output.index("--middle")
        assert help_output.index("--middle") < help_output.index("--zeta")

    def test_sorting_with_no_option_strings(self):
        """
        Test sorting of arguments without option strings.

        Verifies alphabetical ordering of positional arguments.
        """
        parser = argparse.ArgumentParser(formatter_class=AlphabeticalHelpFormatter)

        # Add positional arguments
        parser.add_argument("zeta", help="Zeta argument")
        parser.add_argument("alpha", help="Alpha argument")
        parser.add_argument("middle", help="Middle argument")

        # Capture help output
        help_output = parser.format_help()
        help_output = help_output[help_output.index("\n") :]  # skip first line

        # Verify alphabetical ordering
        assert help_output.index("alpha") < help_output.index("middle")
        assert help_output.index("middle") < help_output.index("zeta")
